"""
Set-up function. Initiates the communication interface.
"""
import json
import os
import sys
from typing import Union

import serial
import serial.tools.list_ports

from tmtc_core.comIF.obsw_com_interface import CommunicationInterface
from tmtc_core.comIF.obsw_dummy_com_if import DummyComIF
from tmtc_core.comIF.obsw_ethernet_com_if import EthernetComIF
from tmtc_core.comIF.obsw_serial_com_if import SerialComIF, SerialCommunicationType
from tmtc_core.comIF.obsw_qemu_com_if import QEMUComIF

from tmtc_core.utility.obsw_logger import get_logger
from tmtc_core.utility.obsw_tmtc_printer import TmTcPrinter

import config.tmtcc_config as g

LOGGER = get_logger()


def set_communication_interface(tmtc_printer: TmTcPrinter) -> Union[CommunicationInterface, None]:
    """
    Return the desired communication interface object
    :param tmtc_printer: TmTcPrinter object.
    :return: CommunicationInterface object
    """
    try:
        if g.G_COM_IF == g.ComInterfaces.Ethernet:
            communication_interface = EthernetComIF(
                tmtc_printer=tmtc_printer, tm_timeout=g.G_TM_TIMEOUT,
                tc_timeout_factor=g.G_TC_SEND_TIMEOUT_FACTOR,
                send_address=g.G_ETHERNET_SEND_ADDRESS,
                receive_address=g.G_ETHERNET_RECV_ADDRESS)
        elif g.G_COM_IF == g.ComInterfaces.Serial:
            serial_baudrate = g.G_SERIAL_BAUDRATE
            serial_timeout = g.G_SERIAL_TIMEOUT
            # Determine COM port, either extract from JSON file or ask from user.
            com_port = determine_com_port()
            communication_interface = SerialComIF(
                tmtc_printer=tmtc_printer, com_port=com_port, baud_rate=serial_baudrate,
                serial_timeout=serial_timeout,
                ser_com_type=SerialCommunicationType.DLE_ENCODING)
            communication_interface.set_dle_settings(
                g.G_SERIAL_DLE_MAX_QUEUE_LEN, g.G_SERIAL_DLE_MAX_FRAME_SIZE, serial_timeout)
        elif g.G_COM_IF == g.ComInterfaces.QEMU:
            serial_timeout = g.G_SERIAL_TIMEOUT
            communication_interface = QEMUComIF(
                tmtc_printer=tmtc_printer,
                serial_timeout=serial_timeout,
                ser_com_type=SerialCommunicationType.DLE_ENCODING)
            communication_interface.set_dle_settings(
                g.G_SERIAL_DLE_MAX_QUEUE_LEN, g.G_SERIAL_DLE_MAX_FRAME_SIZE, serial_timeout)
        else:
            communication_interface = DummyComIF(tmtc_printer=tmtc_printer)
        if not communication_interface.valid:
            LOGGER.warning("Invalid communication interface!")
            sys.exit()
        communication_interface.initialize()
        return communication_interface
    except (IOError, OSError):
        LOGGER.error("Error setting up communication interface")
        LOGGER.exception("Error")
        sys.exit()


def determine_com_port() -> str:
    reconfigure_com_port = False
    com_port = ""

    if os.path.isfile("config/tmtcc_config.json"):
        with open("config/tmtcc_config.json", "r") as write:
            load_data = json.load(write)
            com_port = load_data["COM_PORT"]
            if not check_port_validity(com_port):
                reconfigure = input(
                    "COM port from configuration file not contained within serial"
                    "port list. Reconfigure serial port? [y/n]: ")
                if reconfigure.lower() in ['y', "yes"]:
                    write.close()
                    os.remove("config/tmtcc_config.json")
                    reconfigure_com_port = True
    else:
        reconfigure_com_port = True

    if reconfigure_com_port:
        com_port = prompt_com_port()
        save_to_json = input("Do you want to store serial port to "
                             "configuration? (y/n): ")
        if save_to_json.lower() in ['y', "yes"]:
            with open("config/tmtcc_config.json", "w") as write:
                json.dump(dict(COM_PORT=com_port), write, indent=4)
    return com_port


def prompt_com_port() -> str:
    while True:
        com_port = input(
            "Configuring serial port. Please enter COM Port"
            "(enter h to display list of COM ports): ")
        if com_port == 'h':
            ports = serial.tools.list_ports.comports()
            for port, desc, hwid in sorted(ports):
                print("{}: {} [{}]".format(port, desc, hwid))
        else:
            if not check_port_validity(com_port):
                print("Serial port not in list of available serial ports. Try again? [y/n]")
                try_again = input()
                if try_again.lower() in ['y', "yes"]:
                    continue
                else:
                    break
            else:
                break
    return com_port


def check_port_validity(com_port_to_check: str) -> bool:
    port_list = []
    ports = serial.tools.list_ports.comports()
    for port, desc, hwid in sorted(ports):
        port_list.append(port)
    if com_port_to_check not in port_list:
        return False
    return True

