# -*- coding: utf-8 -*-
"""
Program: obsw_tm_service_3.py
Date: 30.12.2019
Description: Deserialize Housekeeping TM
Author: R. Mueller
"""

from tmtc_core.pus_tm.tmtcc_pus_tm_base import PusTelemetry
from typing import Type
from tmtc_core.utility.obsw_logger import get_logger
import struct
import config.tmtcc_config as g

LOGGER = get_logger()


class Service3TM(PusTelemetry):
    """
    @brief  This class encapsulates the format of Service 3 telemetry
    """
    # Minimal packet contains SID, which consists of object ID(4) and set ID(4)
    MINIMAL_PACKET_SIZE = 8
    HK_START_IDX = MINIMAL_PACKET_SIZE
    # Minimal structure report contains SID (8), reporting status(1), validity flag (1),
    # collection interval as float (4) and number of parameters(1)
    STRUCTURE_REPORT_FIXED_HEADER_SIZE = MINIMAL_PACKET_SIZE + 7

    def __init__(self, byte_array: bytearray):
        super().__init__(byte_array)
        if len(self._tm_data) < 8:
            warning = "Service3TM: handle_filling_definition_arrays: Invalid Service 3 packet," \
                      " is too short!"
            LOGGER.warning(warning)
            return

        self.object_id = struct.unpack('!I', self._tm_data[0:4])[0]
        self.object_id_raw = self._tm_data[0:4]
        self.set_id = struct.unpack('!I', self._tm_data[4:8])[0]
        self.hk_header = []
        self.hk_content = []
        self.hk_definition_header = []
        self.hk_definition = []
        self.number_of_parameters = 0
        self.validity_buffer = []
        self.specify_packet_info("Housekeeping Packet")
        self.param_length = 0
        if self.get_subservice() == 10 or self.get_subservice() == 12:
            self.handle_filling_definition_arrays()
        if self.get_subservice() == 25 or self.get_subservice() == 26:
            self.handle_filling_hk_arrays()

    def append_telemetry_content(self, array):
        super().append_telemetry_content(array)
        array.append(hex(self.object_id))
        array.append(hex(self.set_id))
        array.append(int(self.param_length))

    def append_telemetry_column_headers(self, array):
        super().append_telemetry_column_headers(array)
        array.append("Object ID")
        array.append("Set ID")
        array.append("HK Data Size")

    def handle_filling_definition_arrays(self):
        if len(self._tm_data) < self.STRUCTURE_REPORT_FIXED_HEADER_SIZE:
            warning = "Service3TM: handle_filling_definition_arrays: Invalid structure report " \
                      "from " + str(hex(self.object_id)) + ", is shorter than" + \
                      str(self.STRUCTURE_REPORT_FIXED_HEADER_SIZE) + "."
            LOGGER.warning(warning)
            return
        self.hk_header = ["Object ID", "Set ID", "Report Status", "Is valid",
                          "Collection Interval (s)", "Number Of IDs"]
        reporting_enabled = self._tm_data[8]
        set_valid = self._tm_data[9]
        collection_interval_seconds = struct.unpack('>f', self._tm_data[10:14])[0] / 1000.0
        num_params = self._tm_data[14]
        if len(self._tm_data) < self.STRUCTURE_REPORT_FIXED_HEADER_SIZE + num_params * 4:
            warning = "Service3TM: handle_filling_definition_arrays: Invalid structure report " \
                      "from " + str(hex(self.object_id)) + ", is shorter than " + \
                      str(self.STRUCTURE_REPORT_FIXED_HEADER_SIZE + num_params * 4) + "."
            LOGGER.warning(warning)
            return

        parameters = []
        counter = 1
        for array_index in range(self.STRUCTURE_REPORT_FIXED_HEADER_SIZE,
                                 self.STRUCTURE_REPORT_FIXED_HEADER_SIZE + 4 * num_params, 4):
            parameter = struct.unpack('>I', self._tm_data[array_index:array_index + 4])[0]
            self.hk_header.append("Pool ID " + str(counter))
            parameters.append(str(hex(parameter)))
            counter = counter + 1
        if reporting_enabled == 1:
            status_string = "On"
        else:
            status_string = "Off"
        if set_valid:
            valid_string = "Yes"
        else:
            valid_string = "No"
        self.hk_content = [hex(self.object_id), self.set_id, status_string, valid_string,
                           collection_interval_seconds, num_params]
        self.hk_content.extend(parameters)

    def handle_filling_hk_arrays(self):
        self.param_length = len(self._tm_data) - 8
        # TODO: This can be automated by using the MIB parser pool names and pool datatypes
        if self.object_id_raw == g.THERMAL_SENSOR_DEVICE_ID:
            self.handle_thermal_sensor_hk_data()
        elif self.object_id_raw == g.GYRO_DEVICE_ID:
            self.handle_gyro_hk_data()
        elif self.object_id_raw == g.GPS0_DEVICE_ID or self.object_id == g.GPS1_DEVICE_ID:
            self.handle_gps_hk_data()
        elif self.object_id_raw == g.TEST_TASK_ID:
            self.handle_test_hk_data()
        elif self.object_id_raw == g.INTERNAL_ERROR_REPORTER_ID:
            self.handle_int_err_rprt_data()
        else:
            LOGGER.info("Service 3 TM: Parsing for this SID has not been implemented.")

    def handle_gps_hk_data(self):
        # TODO: size check. sth not right with gps 0 test
        self.number_of_parameters = 9
        self.hk_header = ["Fix Mode", "SV in Fix", "GNSS Week", "Time of Week", "Latitude",
                          "Longitude", "Mean Sea Altitude", "Position X", "Position Y",
                          "Position Z", "Velocity X", "Velocity Y", "Velocity Z"]
        fix_mode = self._tm_data[4]
        sv_in_fix = self._tm_data[5]
        gnss_week = struct.unpack('>H', self._tm_data[6:8])[0]
        time_of_week = struct.unpack('>I', self._tm_data[8:12])[0]
        latitude = struct.unpack('>I', self._tm_data[12:16])[0]
        longitude = struct.unpack('>I', self._tm_data[16:20])[0]
        msa = struct.unpack('>I', self._tm_data[20:24])[0]
        position_x = struct.unpack('>d', self._tm_data[24:32])[0]
        position_y = struct.unpack('>d', self._tm_data[32:40])[0]
        position_z = struct.unpack('>d', self._tm_data[40:48])[0]
        vx = struct.unpack('>d', self._tm_data[48:56])[0]
        vy = struct.unpack('>d', self._tm_data[56:64])[0]
        vz = struct.unpack('>d', self._tm_data[64:72])[0]
        self.hk_content = [fix_mode, sv_in_fix, gnss_week, time_of_week, latitude, longitude, msa,
                           position_x, position_y, position_z, vx, vy, vz]
        self.validity_buffer = self._tm_data[72:]
        # print(self.validityBuffer)
        # print(str(format(self.validityBuffer[0], '#010b')))
        # print(str(format(self.validityBuffer[1], '#010b')))
        # print("Validity Buffer Length: " + str(len(self.validityBuffer)))

    def handle_test_hk_data(self):
        self.number_of_parameters = 6
        self.hk_header = ["Bool", "UINT8", "UINT16", "UINT32", "FLOAT1", "FLOAT2"]
        test_bool = self._tm_data[4]
        test_uint8 = self._tm_data[5]
        test_uint16 = (self._tm_data[6] << 8) | self._tm_data[7]
        test_uint32 = struct.unpack('>I', self._tm_data[8:12])[0]
        float_vector1 = struct.unpack('>f', self._tm_data[12:16])[0]
        float_vector2 = struct.unpack('>f', self._tm_data[16:20])[0]
        self.hk_content = [test_bool, test_uint8, test_uint16,
                           test_uint32, float_vector1, float_vector2]
        self.validity_buffer = self._tm_data[20:]
        # print(self.validityBuffer)
        # print(str(format(self.validityBuffer[0], '#010b')))
        # print("Validity Buffer Length: " + str(len(self.validityBuffer)))

    def handle_thermal_sensor_hk_data(self):
        self.number_of_parameters = 2
        if len(self._tm_data) < 13:
            LOGGER.error("Format of thermal sensor HK data might be invalid!")
            return

        self.hk_header = ["Temperature", "Error Byte"]
        temperature_celcius = struct.unpack('!f', self._tm_data[8:12])[0]
        error_byte = self._tm_data[12]
        self.hk_content = [temperature_celcius, error_byte]
        self.validity_buffer = self._tm_data[13:]

    def handle_gyro_hk_data(self):
        self.number_of_parameters = 3
        if len(self._tm_data) < 21:
            LOGGER.error("Format of thermal sensor HK data might be invalid!")
            return

        self.hk_header = ["AngVel X", "Ang Vel Y", "AngVel Z"]
        angular_velocity_x = struct.unpack('!f', self._tm_data[8:12])[0]
        angular_velocity_y = struct.unpack('!f', self._tm_data[12:16])[0]
        angular_velocity_z = struct.unpack('!f', self._tm_data[16:20])[0]
        self.hk_content = [angular_velocity_x, angular_velocity_y, angular_velocity_z]
        self.validity_buffer = self._tm_data[20:]

    def handle_int_err_rprt_data(self):
        self.number_of_parameters = 3
        if len(self._tm_data) < 12:
            LOGGER.error("Format of thermal sensor HK data might be invalid!")
            return
        self.hk_header = ["TM Hits", "Queue Hits", "Store Hits"]
        tm_hits = struct.unpack('!I', self._tm_data[self.HK_START_IDX:self.HK_START_IDX + 4])[0]
        queue_hits = struct.unpack(
            '!I', self._tm_data[self.HK_START_IDX + 4:self.HK_START_IDX + 8])[0]
        store_hits = struct.unpack(
            '!I', self._tm_data[self.HK_START_IDX + 8:self.HK_START_IDX + 12])[0]
        self.hk_content = [tm_hits, queue_hits, store_hits]
        self.validity_buffer = self._tm_data[self.HK_START_IDX + 12:]


Service3TM: Type[PusTelemetry]
