#!/usr/bin/python3.8
"""
@brief      Binary Uploader Module
@details
This module will be used to upload binaries to the OBC via a communication port, given
a supplied binary. The binary will be sent via the specified communication interface.
It will be possible to encode the data (for example using DLE encoding)
"""
import os
import time
import tkinter as tk
from tkinter import filedialog
from collections import deque
from glob import glob

from tmtc_core.pus_tm.tmtcc_tm_service_1 import Service1TM
from tmtc_core.comIF.obsw_com_interface import CommunicationInterface
from tmtc_core.pus_tc.tmtcc_pus_tc_base import TcQueueT, TcDictionaryKeys
from utility.tmtcc_file_transfer_helper import FileTransferHelper
import config.tmtcc_config as g
from tmtc_core.utility.obsw_tmtc_printer import TmTcPrinter, DisplayMode
from tmtc_core.utility.obsw_logger import get_logger
from tmtc_core.sendreceive.obsw_tm_listener import TmListener

LOGGER = get_logger()


class BinaryFileUploader:
    def __init__(self, com_if: CommunicationInterface, tmtc_printer: TmTcPrinter,
                 tm_listener: TmListener):
        """
        Initializes the binary file uploader with the required components.
        @param com_if:
        @param tmtc_printer:
        @param tm_listener:
        """
        self.com_if = com_if
        self.tmtc_printer = tmtc_printer
        self.tm_listener = tm_listener
        self.iobc = False
        # Will be set later depending on board.
        self.send_interval = 0

    def perform_file_upload(self):
        gui_cl_prompt = input("GUI(0) or command line version (1)? [0/1]: ")
        if gui_cl_prompt == 0:
            gui_cl_prompt = True
        else:
            gui_cl_prompt = False
        print("Please select file to upload: ")
        file_path = ""
        if gui_cl_prompt:
            root = tk.Tk()
            root.withdraw()
            root.wm_attributes('-topmost', 1)
            file_path = filedialog.askopenfilename(parent=root)
            print("File select: " + str(file_path))
            if file_path == ():
                LOGGER.warning("Invalid file path, exiting binary upload mode.")
                return
        else:
            result = [y for x in os.walk("../_bin") for y in glob(os.path.join(x[0], '*.bin'))]
            print("Files found in _bin folder: ")
            for idx, path in enumerate(result):
                print("Selection " + str(idx) + ": " + str(path))
            select_valid = False
            selection = input("Please enter desired selection [c to cancel]: ")
            while not select_valid:
                if selection == 'c':
                    print("Exiting binary upload mode..")
                    return
                if not selection.isdigit():
                    selection = input("Invalid input, try again [c to cancel]: ")
                if selection.isdigit():
                    if int(selection) < len(result):
                        file_path = result[int(selection)]
                        select_valid = True
                    else:
                        selection = input("Invalid input, try again [c to cancel]: ")

        print_string = file_path.rsplit(os.path.sep, 1)[-1] + " was selected."
        LOGGER.info(print_string)
        calc_hamming_code = input("Calculate and send hamming code? [y/n]: ")
        if calc_hamming_code in ['y', 'yes', 1]:
            calc_hamming_code = True
            print("Hamming code will be calculated and sent in tail packet")
        else:
            calc_hamming_code = False
            print("Hamming code will not be calculated")

        iobc_prompt = input("iOBC? [y/n]: ")
        if iobc_prompt in ['y', 'yes', 1]:
            self.iobc = True
            self.send_interval = 1.4
            iobc_prompt = True
        else:
            self.iobc = False
            self.send_interval = 0.6
            iobc_prompt = False

        bootloader_prompt = input("Bootloader (0) or Software Image (1)? [0/1]: ")
        if str(bootloader_prompt) == "0":
            bootloader_prompt = True
        else:
            bootloader_prompt = False

        prompt_lock = input("Lock file with last packet? [y/n]: ")
        if prompt_lock in ['n', "no", 0]:
            prompt_lock = False
        else:
            prompt_lock = True

        if bootloader_prompt:
            file_name = "bl.bin"
        else:
            file_name = "obsw_up.bin"

        if iobc_prompt:
            if bootloader_prompt:
                repository_name = "BIN/IOBC/BL"
            else:
                repository_name = "BIN/IOBC/OBSW"
        else:
            if bootloader_prompt:
                repository_name = "BIN/AT91/BL"
            else:
                repository_name = "BIN/AT91/OBSW"

        if calc_hamming_code:
            pass

        # Right now, the size of PUS packets is limited to 1500 bytes which also limits the app
        # data length.
        frame_length = g.G_MAX_APP_DATA_LENGTH

        if calc_hamming_code:
            # now we calculate the hamming code
            pass

        tc_queue = deque()

        # Delete existing binary file first, otherwise data might be appended to otherwise
        # valid file which already exists.
        file_transfer_helper = FileTransferHelper(
            tc_queue=tc_queue, max_size_of_app_data=frame_length, target_repository=repository_name,
            target_filename=file_name)

        init_ssc = 0
        self.tmtc_printer.set_display_mode(DisplayMode.SHORT)

        # Configure file transfer helper
        file_transfer_helper.set_data_from_file(file_path)
        file_transfer_helper.set_to_delete_old_file()
        if prompt_lock:
            file_transfer_helper.set_to_lock_file(prompt_lock)
        else:
            file_transfer_helper.set_to_lock_file(prompt_lock)

        # Generate the packets.
        file_transfer_helper.generate_packets(init_ssc)

        self.tm_listener.set_listener_mode(TmListener.ListenerModes.MANUAL)
        print_string = "BinaryUploader: Detected file size: " + str(
            file_transfer_helper.file_size())
        LOGGER.info(print_string)
        total_num_packets = file_transfer_helper.get_number_of_packets_generated()
        print_string = "BinaryUploader: " + str(total_num_packets) + \
                       " packets generated."
        if prompt_lock:
            print_string += " File will be locked."
        else:
            print_string += " File will not be locked."
        LOGGER.info(print_string)

        packets_received = self.__perform_send_algorithm(tc_queue, total_num_packets)

        LOGGER.info("BinaryUploader: All binary packets were sent!")
        print_string = str(packets_received) + " replies received."

        LOGGER.info(print_string)
        self.tm_listener.clear_tm_packet_queue()
        LOGGER.info("Upload operation finished successfully.")

    def __perform_send_algorithm(self, tc_queue: TcQueueT, number_of_packets: int) -> int:
        last_check = time.time()
        last_sent = time.time()
        reception_dict = dict()
        total_time = self.send_interval * number_of_packets
        idx = 1
        new_packets_received = 0
        num_packets_received = 0
        while tc_queue:
            next_send = last_sent + self.send_interval
            tc_packet, tc_info = tc_queue.pop()
            if not isinstance(tc_packet, str):
                # print_string = "Sending packet " + str(idx) + ".."
                # LOGGER.info(print_string)
                idx += 1
                self.com_if.send_telecommand(tc_packet, tc_info)
                self.tmtc_printer.print_telecommand(tc_packet, tc_info)
                # Store a list which tracks whether acceptance messages were received
                # The first entry is a simply counter.
                reception_dict.update({tc_info[TcDictionaryKeys.SSC]: [0, False, False, False]})
            elif tc_packet == "print":
                LOGGER.info(tc_info)
            remaining_time_string = "Remaining time: " + \
                                    str(round(total_time - (idx - 2) * self.send_interval, 2)) + \
                                    " seconds"
            print_progress_bar(idx - 2, number_of_packets, print_end="\n",
                               suffix=remaining_time_string)
            # sys.stdout.write("\033[F")  # Cursor up one line
            new_packets_received += self.__handle_tm_queue(reception_dict)

            # Every 5 seconds, check whether any reply has been received. If not, cancel operation.
            if time.time() - last_check > 5.0 and new_packets_received == 0:
                LOGGER.warning("No replies are being received, cancelling upload operation..")
            elif time.time() - last_check > 5.0:
                num_packets_received += new_packets_received
                new_packets_received = 0
                last_check = time.time()

            time_to_sleep = next_send - time.time()
            last_sent = next_send
            time.sleep(time_to_sleep)

        # handle last telemetry packets coming in.
        wait_time = 4
        LOGGER.info("Waiting " + str(wait_time) + " more seconds for TM packets..")
        time_till_wake = time.time() + wait_time
        while time.time() < time_till_wake:
            new_packets_received += self.__handle_tm_queue(reception_dict)
            time.sleep(1)

        num_packets_received += new_packets_received
        return num_packets_received

    def __handle_tm_queue(self, reception_dict: dict) -> int:
        num_packets_received = 0
        packets_received = self.tm_listener.retrieve_tm_packet_queue()
        self.tm_listener.clear_tm_packet_queue()
        # reception_deque.extend(packets_received)
        for packet_list in packets_received:
            for packet in packet_list:
                num_packets_received += 1
                if packet.get_service() == 1:
                    packet: Service1TM
                    tc_ssc = packet.get_tc_ssc()
                    acceptance_list = reception_dict.get(tc_ssc)
                    if acceptance_list is None:
                        LOGGER.warning("Invalid TC SSC with number " + str(tc_ssc) +
                                       " detected!")
                        continue
                    if packet.get_subservice() == 1:
                        acceptance_list[1] = True
                    if packet.get_subservice() == 3:
                        acceptance_list[2] = True
                    if packet.get_subservice() == 7:
                        acceptance_list[3] = True
                    reception_dict.update({tc_ssc: acceptance_list})
                elif packet.get_service() == 5:
                    # TODO: print event
                    LOGGER.info("Event received!")
                else:
                    print_string = "Other TM packet with service ID " + str(packet.get_service()) + \
                                   " received!"
                    LOGGER.info(print_string)
        clear_list = []
        for key, reception_list in reception_dict.items():
            if reception_list[1] and reception_list[2] and reception_list[3]:
                # All replies received, we can delete the entry and confirm succesfull handling
                print_string = "All replies received for upload packet with SSC " + str(key) + "."
                LOGGER.info(print_string)
                clear_list.append(key)
            elif reception_list[0] > 5:
                # No replies for telecommand.
                LOGGER.warning("No reply received for TC SSC " + str(key) + "!")
                # Continue nonetheless
                clear_list.append(key)
        for key in clear_list:
            del reception_dict[key]
        return num_packets_received


# https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
# Thank you Greensticks :-)
def print_progress_bar(iteration, total, prefix='', suffix='', decimals=1, length=100,
                       fill='█', print_end="\r"):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        print_end    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filled_length = int(length * iteration // total)
    bar = fill * filled_length + '-' * (length - filled_length)
    print(f'\r{prefix} |{bar}| {percent}% {suffix}', end=print_end)
    # Print New Line on Complete
    if iteration == total:
        print()
